%{
/*
 * exportfs4.y
 *
 *  Copyright (c) 2000 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Jake Moilanen <jmoilane@umich.edu>
 *  Kendrick Smith <kmsmith@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include "exportfs4.h"
#include "pseudodev.h"
  
#define YYDEBUG 1
int yydebug = 0;

static int grace_period;
%}

%union {
  char *element;
}

%token PATH ELEMENT END_ELEMENT NL RO SEC
%type<element> entry v_path e_path f_perm

%%
  
finish: tree                       { arg.ca_start.grace_period = grace_period; 
                                     start_server();}

tree:   /* empty */
| entry
| entry NL tree
| NL tree
;

entry: v_path e_path             { add_export(); reset_struct(); } 
| v_path e_path f_perm           { add_export(); reset_struct(); }
| v_path e_path f_perm f_sec     { add_export(); reset_struct(); }
| v_path e_path f_sec            { add_export(); reset_struct(); }

v_path: END_ELEMENT {} 
|    ELEMENT v_path {} ;             

e_path: PATH {}

f_perm: RO {}

f_sec: SEC {}
|    SEC f_sec {};

%%

extern FILE *yyin;

int
main (int argc, char **argv)
{
  struct stat s;
  int i;
  char c;
  int err;
  
  /* init */
  grace_period = 0;

  err = open_pseudodev();
  if (err) {
    fprintf(stderr, "couldn't open /etc/nfs4-pseudofs: %s\n", strerror(err));
    exit(1);
  }

  /* command line checking */
  if (argc > 1){
    while ((c = getopt(argc, argv, "gsu:")) != EOF) {
      switch (c) {
      case 'u':
	err = unexport(optarg);
	if (err) {
	  fprintf(stderr, "unexport: %s\n", strerror(err));
	  exit(1);
	}
	exit(0);
      case 's':
	destroy_server();
	break;
      case 'g': 
	grace_period = 1;
	break;
      default:
	printf("exportfs4: invalid option -- %c\n", c);
	printf("usage: exportfs4 [-g|-s|-u filename]\n");
      break;
      } // switch
    } // while
  } // if
  
  if (stat(EXPORT_FILE, &s)){
    perror("stat: File probably doesn't exist\n");
    exit(1);
  }
  if (s.st_uid != 0){
    perror("Export file not owned by root\n");
    exit(1);
  }

  yyin = fopen(EXPORT_FILE, "r");
  if (!yyin) {
    fprintf(stderr,"could not open %s\n", EXPORT_FILE);
    exit(1);
  }

  err = unexport_all();
  if (err) {
    perror("when attempting to remove old exports");
    exit(1);
  }
  
  while(!feof(yyin)) {  
    yyparse();
  }
  exit(0);
}

