/*
 * util.c 
 *
 *  Copyright (c) 2000 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Jake Moilanen <jmoilane@umich.edu>
 *  Kendrick Smith <kmsmith@umich.edu>
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "exportfs4.h"
#include "pseudodev.h"

void
add_pseudo(void)
{
  int result;

  memset(&arg, 0, sizeof(arg));
  arg.ca_version = NFSCTL_VERSION;
  arg.ca_addpseudo.namlen = strlen(components[ncomponents - 1]);
  arg.ca_addpseudo.name = components[ncomponents - 1];
  
  result = get_pseudodev(ncomponents - 1, components, &arg.ca_addpseudo.parent_major,
		      &arg.ca_addpseudo.parent_minor, &arg.ca_addpseudo.parent_ino);
  if (result) {
    fprintf(stderr, "couldn't access pseudofs: %s\n", strerror(result));
    destroy_server();
  }

  result = get_pseudodev(ncomponents, components, &arg.ca_addpseudo.major,
		      &arg.ca_addpseudo.minor, &arg.ca_addpseudo.ino);
  if (result) {
    fprintf(stderr, "couldn't access pseudofs: %s\n", strerror(result));
    destroy_server();
  }

  result = nfsservctl(NFS4CTL_ADDPSEUDO, &arg, &res);
  if (result){
    perror("addpseudo");
    destroy_server();
  }
}

void
add_export(void)
{
  struct stat s;
  int result;

  arg.ca_version = NFSCTL_VERSION;

  if (stat(arg.ca_export.path, &s)){
    perror("stat");
    exit(1);
  }

  printf("AAA add_export: ncomponents=%d\n", ncomponents);
  result = get_pseudodev(ncomponents, components, &arg.ca_export.major,
			 &arg.ca_export.minor, &arg.ca_export.ino);
  if (result) {
    fprintf(stderr, "couldn't access pseudofs: %s\n", strerror(result));
    destroy_server();
  }
  
  result = nfsservctl(NFS4CTL_EXPORT, &arg, &res);
  if (result){
    perror(arg.ca_export.path);
    destroy_server();
  }
}

/*
 * reset_struct
 *
 * This is mainly to reset the permission flags like read only and 
 * the security flags.
 *
 */
void 
reset_struct(void)
{
  int i;
  
  if (arg.ca_export.path) {
    free(arg.ca_export.path);
    arg.ca_export.pathlen = 0;
    arg.ca_export.path = NULL;
  }

  for (i = 0; i < ncomponents; i++)
    free(components[i]);
  ncomponents = 0;
  
  /* XXX: why not just memset(&arg, 0, sizeof(arg)); at this point? */

  /* file permission flags */
  arg.ca_export.read_only = 0;

  /* Sec flags */
  arg.ca_export.sec_none = 0;
  arg.ca_export.sec_sys = 0; 
  arg.ca_export.sec_dh = 0;
  arg.ca_export.sec_krb5 = 0;
  arg.ca_export.sec_krb5i = 0;
  arg.ca_export.sec_krb5p = 0;
  arg.ca_export.sec_spkm3 = 0; 
  arg.ca_export.sec_spkm3i = 0;
  arg.ca_export.sec_spkm3p = 0;
  arg.ca_export.sec_lkey = 0;
  arg.ca_export.sec_lkeyi = 0;
  arg.ca_export.sec_lkeyp = 0;
}

void
start_server(void)
{
  arg.ca_version = NFSCTL_VERSION;
  arg.ca_start.nthreads = NFSD4_NTHREADS;

  if (nfsservctl(NFS4CTL_START, &arg, &res)) {
    perror("export");
    destroy_server();
  }
}

int
unexport_all(void)
{
  memset(&arg, 0, sizeof(arg));
  arg.ca_version = NFSCTL_VERSION;
  if (nfsservctl(NFS4CTL_REMOVEPSEUDO, &arg, &res))
    return errno;
  return 0;
}

void
destroy_server(void)
{
  unexport_all();

  memset(&arg, 0, sizeof(arg));
  arg.ca_version = NFSCTL_VERSION;
  arg.ca_start.nthreads = 0;
  nfsservctl(NFS4CTL_START, &arg, &res);
  exit(1);
}

int
unexport(char *filename)
{
  memset(&arg, 0, sizeof(arg));
  arg.ca_version = NFSCTL_VERSION;
  arg.ca_unexport.namelen = strlen(filename);
  arg.ca_unexport.name = filename;
  if (nfsservctl(NFS4CTL_UNEXPORT, &arg, &res))
    return errno;
  return 0;
}
